/**
 * @file Photoresistor_Tracker_Basic.ino
 * @brief Basic Light Tracker using Photoresistors and AccelStepper for Arduino Uno.
 * @version 0.5 (Basic Adaptation)
 * @date 2025-04-28 (Original base date)
 *
 * @details This firmware controls a 2-axis stepper motor system (like a pan-tilt
 * mechanism) to track a light source using five photoresistors (L, R, U, D, C).
 * It includes homing, searching (sweep, basic local), tracking (P-control), and error handling.
 * Simplified from an IR tracker, focusing on basic functionality.
 *
 * @warning CRITICAL: Verify all pin assignments in the configuration section match
 *          your hardware EXACTLY before uploading. Limit switch pins have changed!
 * @warning CRITICAL: Photoresistor pins A0-A4 are used. If hardware I2C (which often uses A4/A5)
 *          is required, photoresistor pin assignments MUST change, or use software I2C.
 */

#include <Arduino.h>
#include <AccelStepper.h>
#include <limits.h>
#include <avr/pgmspace.h>

// ==========================================================================
// ===                  SYSTEM CONFIGURATION & TUNING                     ===
// ==========================================================================
// --- System Version ---
#define SYSTEM_VERSION "0.5-Photoresistor-Basic"

// --- Build Configuration ---
// Comment out the next line for a production build (disables most Serial output for speed)
#define DEBUG_BUILD

// Derived Conditional Compilation
#ifdef DEBUG_BUILD
#define ENABLE_TELEMETRY	  // Enable periodic status messages via Serial
#define ENABLE_DEBUG_PRINTING // Enable state change and event messages via Serial
#else
// Production build: Telemetry and Debug prints are disabled below
#undef ENABLE_TELEMETRY
#undef ENABLE_DEBUG_PRINTING
#endif

// --- Pin Definitions ---
// *** VERIFY ALL PINS AGAINST YOUR HARDWARE ***
// Stepper Motors (using AccelStepper HALF4WIRE mode)
const uint8_t stepperX_pin1 = 8;
const uint8_t stepperX_pin3 = 9;
const uint8_t stepperX_pin2 = 10;
const uint8_t stepperX_pin4 = 11;

const uint8_t stepperY_pin1 = 4;
const uint8_t stepperY_pin3 = 5;
const uint8_t stepperY_pin2 = 6;
const uint8_t stepperY_pin4 = 7;

// Photoresistors (Analog Inputs)
// Assumes higher analogRead() value = more light.
const uint8_t photoresistorPinL = A0; // Left
const uint8_t photoresistorPinR = A1; // Right
const uint8_t photoresistorPinU = A2; // Up
const uint8_t photoresistorPinD = A3; // Down
const uint8_t photoresistorPinC = A4; // Center

// Limit Switches (Active LOW assumed) - MOVED TO DIGITAL PINS
const uint8_t limitSwitchPinX = 2; // PD2 (INT0) - Connect X-axis limit switch here
const uint8_t limitSwitchPinY = 3; // PD3 (INT1) - Connect Y-axis limit switch here

// --- Photoresistor & Tracking Parameters (CRITICAL TUNING REQUIRED) ---
const int PHOTO_LIGHT_THRESHOLD_GENERAL = 300; // Min analog reading to consider any sensor "active" (0-1023)
const int PHOTO_LIGHT_THRESHOLD_CENTER = 400;  // Min analog reading for Center sensor to influence logic (if used beyond general detection)
// Defines how much raw analog difference translates into an "error unit" for Kp.
// Larger value = less sensitive to small differences. (e.g. (L_val - R_val) / THIS_FACTOR = error_int)
// Example: If L=600, R=400, diff=200. If factor is 50, error_int = 4. If factor is 100, error_int = 2.
const int ANALOG_DIFF_SCALING_FACTOR = 50; // Tune this based on your LDR sensitivity and light source!

// --- Motor & Movement Parameters ---
const uint16_t stepsPerRevolution = 2038;
const float homingSpeed = 200.0;
const float homingAcceleration = 200.0;
const float sweepSpeedDefault = 300.0; // Reduced default for initial basic setup
const float sweepAccelerationDefault = 200.0;
const float trackSpeedDefault = 300.0;
const float trackAccelerationDefault = 200.0;
const float localSearchSpeed = 250.0;
const float localSearchAccel = 150.0;

// Search Range & Homing
const long searchRangeX_deg = 180;
const long searchRangeY_deg = 90;
const int postHomingMoveOffSwitchSteps = 30;
const uint8_t HOMING_OVERSHOOT_MULTIPLIER = 20;

// --- Tracking Control (Proportional Controller - REQUIRES CRITICAL TUNING) ---
const float Kp_X_float = 10.0;		  // Proportional gain for X-axis. Start low (e.g., 5-20), tune carefully!
const float Kp_Y_float = 10.0;		  // Proportional gain for Y-axis.
const int trackingDeadband_steps = 8; // Ignore movement commands smaller than this (steps)

// --- Local Search Strategy Parameters ---
const int localSearchProbeSteps = 25;
const int localSearchSweepSteps = 80;
const int localSearchBoxStepSize = 25;
const uint8_t localSearchMaxCycles = 2; // Reduced for basic
const uint8_t LOCAL_SEARCH_BOX_STEPS_PER_LAYER = 8;

// --- Timing Parameters ---
const unsigned int debounceDelayMs = 50; // General sensor debounce (less critical for LDRs but kept for limit switches)
const unsigned int limitSwitchDebounceMs = 25;
const unsigned int signalLostTimeoutMs = 700;
const unsigned int searchFailDelayMs = 3000;
const unsigned long stateTimeoutMs = 25000;
const unsigned long homingTimeoutMs = 45000;
const unsigned long localSearchTimeoutMs = 15000;
const unsigned int telemetryIntervalMs = 750;
const unsigned int postHomingPauseMs = 500;

// --- Stall Detection Parameters ---
const unsigned int stallCheckIntervalMs = 100;
const unsigned int stallTimeoutThresholdMs = 1800;
const uint8_t stallPositionTolerance = 2;

// ==========================================================================
// ===                   INTERNAL DEFINITIONS & GLOBALS                   ===
// ==========================================================================
const long searchMaxX_steps = (long)(((float)searchRangeX_deg * stepsPerRevolution) / 360.0f);
const long searchMaxY_steps = (long)(((float)searchRangeY_deg * stepsPerRevolution) / 360.0f);
const long searchMinX_steps = 0;
const long searchMinY_steps = 0;

#define SENSOR_L 0
#define SENSOR_R 1
#define SENSOR_U 2
#define SENSOR_D 3
#define SENSOR_C 4
const int NUM_SENSORS = 5;

// Array for photoresistor pins for easier iteration
const uint8_t photoresistorPins[NUM_SENSORS] = {
	photoresistorPinL, photoresistorPinR, photoresistorPinU, photoresistorPinD, photoresistorPinC};

// Precomputed Bit Masks for Limit Switches (Direct Port I/O on PORTD)
const uint8_t limitSwitchMaskX = (1 << PIND2); // For limitSwitchPinX on D2
const uint8_t limitSwitchMaskY = (1 << PIND3); // For limitSwitchPinY on D3

enum class State : uint8_t
{
	INITIALIZING,
	CHECK_PIN_CONFLICTS,
	HOMING_X,
	HOMING_Y,
	POST_HOMING_MOVE_OFF_X,
	POST_HOMING_MOVE_OFF_Y,
	POST_HOMING_DELAY,
	SWEEP_X,
	SWEEP_Y,
	CENTERING,
	TRACKING,
	LOCAL_SEARCH,
	SEARCH_FAILED,
	ERROR
};
State currentState = State::INITIALIZING;

const char state0[] PROGMEM = "INITIALIZING";
const char state1[] PROGMEM = "CHECK_PIN_CONFLICTS";
const char state2[] PROGMEM = "HOMING_X";
// ... (rest of state names are the same as original, ensure array matches enum)
const char state3[] PROGMEM = "HOMING_Y";
const char state4[] PROGMEM = "POST_HOMING_MOVE_OFF_X";
const char state5[] PROGMEM = "POST_HOMING_MOVE_OFF_Y";
const char state6[] PROGMEM = "POST_HOMING_DELAY";
const char state7[] PROGMEM = "SWEEP_X";
const char state8[] PROGMEM = "SWEEP_Y";
const char state9[] PROGMEM = "CENTERING";
const char state10[] PROGMEM = "TRACKING";
const char state11[] PROGMEM = "LOCAL_SEARCH";
const char state12[] PROGMEM = "SEARCH_FAILED";
const char state13[] PROGMEM = "ERROR";

const char *const stateNames[] PROGMEM = {
	state0, state1, state2, state3, state4, state5, state6, state7,
	state8, state9, state10, state11, state12, state13};
char stateNameBuffer[25];

AccelStepper stepperX(AccelStepper::HALF4WIRE, stepperX_pin1, stepperX_pin3, stepperX_pin2, stepperX_pin4);
AccelStepper stepperY(AccelStepper::HALF4WIRE, stepperY_pin1, stepperY_pin3, stepperY_pin2, stepperY_pin4);

// Sensor readings and states
int sensorReadings[NUM_SENSORS] = {0};	  // Stores raw analogRead values
bool sensorActive[NUM_SENSORS] = {false}; // True if sensorReading > PHOTO_LIGHT_THRESHOLD_GENERAL
bool limitSwitchActiveX = false;
bool limitSwitchActiveY = false;

// Internal Debounce Tracking for Limit Switches
uint8_t lastRawPinStatesD = 0; // Raw PORTD state for limit switch inputs
unsigned long lastDebounceTimeLimits = 0;

unsigned long lastSignalTime = 0;
unsigned long stateEntryTime = 0;
unsigned long lastTelemetryTime = 0;

int8_t sweepDirectionX = 1;
int8_t sweepDirectionY = 1;
uint8_t sweepXCyclesCompleted = 0;
uint8_t sweepYCyclesCompleted = 0;

long lastKnownX = 0;
long lastKnownY = 0;
int lastErrorX_int = 0; // Stores scaled analog difference
int lastErrorY_int = 0; // Stores scaled analog difference
uint8_t currentLocalSearchPhase = 0;
uint8_t currentLocalSearchCycle = 0;

long lastKnownStepperXPosForStall = 0;
long lastKnownStepperYPosForStall = 0;
unsigned long stallStartTimeX = 0;
unsigned long stallStartTimeY = 0;
unsigned long lastStallCheckTime = 0;

bool recoveryAttempted = false;

// ==========================================================================
// ===                       FUNCTION PROTOTYPES                          ===
// ==========================================================================
bool setupPinsAndCheckConflicts();
void printStartupMessage();
void printTuningGuide();
void readPhotoresistorInputs(); // Changed name
void handleStateMachine();
void changeState(State newState);
const char *getStateName(State state);
inline bool isAnySensorGenerallyActive();	 // Changed name
inline bool isCenterSensorGenerallyActive(); // Changed name
void handleHomingStates();
void handleTrackingLogic(); // Changed name
void handleLocalSearch();
void startSweepX();
void startSweepY();
// void updateStatusLEDsOptimized(); // REMOVED
void checkStateTimeouts();
void checkMotorStall();
void handleSystemError(const __FlashStringHelper *reason);
void indicateFatalError(const __FlashStringHelper *message);
inline unsigned long safeMillisSubtract(unsigned long current, unsigned long previous);
inline long clamp(long val, long minVal, long maxVal);
void setMotorSpeeds(float speedX, float accelX, float speedY, float accelY);
void printTelemetry();

// ==========================================================================
// ===                            SETUP                                   ===
// ==========================================================================
void setup()
{
#if defined(ENABLE_DEBUG_PRINTING) || defined(ENABLE_TELEMETRY)
	Serial.begin(115200);
	delay(2000);
#endif

	printStartupMessage();
	changeState(State::CHECK_PIN_CONFLICTS);
}

// ==========================================================================
// ===                          MAIN LOOP                                 ===
// ==========================================================================
void loop()
{
	unsigned long now = millis();

	stepperX.run();
	stepperY.run();

	readPhotoresistorInputs(); // Reads photoresistors and limit switches

	if ((currentState == State::TRACKING || currentState == State::CENTERING) && isAnySensorGenerallyActive())
	{
		lastSignalTime = now;
	}

	handleStateMachine();
	// updateStatusLEDsOptimized(); // REMOVED
	checkMotorStall();
	checkStateTimeouts();

#ifdef ENABLE_TELEMETRY
	if (currentState > State::CHECK_PIN_CONFLICTS && currentState != State::ERROR &&
		safeMillisSubtract(now, lastTelemetryTime) >= telemetryIntervalMs)
	{
		printTelemetry();
		lastTelemetryTime = now;
	}
#endif
}

// ==========================================================================
// ===                   STATE MACHINE IMPLEMENTATION                     ===
// ==========================================================================
void handleStateMachine()
{
	if (currentState >= State::HOMING_X && currentState <= State::POST_HOMING_DELAY)
	{
		handleHomingStates();
		return;
	}

	bool anySensor = isAnySensorGenerallyActive();
	unsigned long now = millis();

	switch (currentState)
	{
	case State::INITIALIZING:
		changeState(State::CHECK_PIN_CONFLICTS);
		break;

	case State::CHECK_PIN_CONFLICTS:
		if (!setupPinsAndCheckConflicts())
		{
			while (1)
				;
		}
		printTuningGuide();
#ifdef ENABLE_DEBUG_PRINTING
		Serial.println(F("Pin checks complete. Starting Homing sequence..."));
#endif
		recoveryAttempted = false;
		changeState(State::HOMING_X);
		break;

	case State::SWEEP_X:
		if (anySensor)
		{
#ifdef ENABLE_DEBUG_PRINTING
			Serial.println(F("Signal Found (X Sweep) -> Centering"));
#endif
			stepperX.stop();
			recoveryAttempted = false;
			sweepXCyclesCompleted = 0;
			changeState(State::CENTERING);
		}
		else if (stepperX.distanceToGo() == 0)
		{
			sweepXCyclesCompleted++;
			if (sweepXCyclesCompleted >= 2)
			{
#ifdef ENABLE_DEBUG_PRINTING
				Serial.println(F("Full X Sweep Cycle Complete (No Signal) -> Sweeping Y"));
#endif
				sweepXCyclesCompleted = 0;
				sweepDirectionY = 1;
				changeState(State::SWEEP_Y);
			}
			else
			{
#ifdef ENABLE_DEBUG_PRINTING
				Serial.print(F("X Sweep Leg Complete ("));
				Serial.print(sweepXCyclesCompleted);
				Serial.println(F("/2) -> Reversing X Sweep"));
#endif
				sweepDirectionX *= -1;
				startSweepX();
			}
		}
		break;

	case State::SWEEP_Y:
		if (anySensor)
		{
#ifdef ENABLE_DEBUG_PRINTING
			Serial.println(F("Signal Found (Y Sweep) -> Centering"));
#endif
			stepperY.stop();
			recoveryAttempted = false;
			sweepYCyclesCompleted = 0;
			changeState(State::CENTERING);
		}
		else if (stepperY.distanceToGo() == 0)
		{
			sweepYCyclesCompleted++;
			if (sweepYCyclesCompleted >= 2)
			{
#ifdef ENABLE_DEBUG_PRINTING
				Serial.println(F("Full Y Sweep Cycle Complete (No Signal) -> Search FAILED"));
#endif
				sweepYCyclesCompleted = 0;
				changeState(State::SEARCH_FAILED);
			}
			else
			{
#ifdef ENABLE_DEBUG_PRINTING
				Serial.print(F("Y Sweep Leg Complete ("));
				Serial.print(sweepYCyclesCompleted);
				Serial.println(F("/2) -> Reversing Y Sweep"));
#endif
				sweepDirectionY *= -1;
				startSweepY();
			}
		}
		break;

	case State::CENTERING:
		handleTrackingLogic();
		if (stepperX.distanceToGo() == 0 && stepperY.distanceToGo() == 0)
		{
#ifdef ENABLE_DEBUG_PRINTING
			Serial.println(F("Centering Move Complete. Re-evaluating..."));
#endif
			recoveryAttempted = false;
			if (isAnySensorGenerallyActive())
			{
#ifdef ENABLE_DEBUG_PRINTING
				Serial.println(F("Signal Active -> Tracking"));
#endif
				changeState(State::TRACKING);
			}
			else
			{
#ifdef ENABLE_DEBUG_PRINTING
				Serial.println(F("Signal LOST during/after Centering -> Local Search"));
#endif
				changeState(State::LOCAL_SEARCH);
			}
		}
		break;

	case State::TRACKING:
		if (anySensor)
		{
			lastKnownX = stepperX.currentPosition();
			lastKnownY = stepperY.currentPosition();
			handleTrackingLogic();
			recoveryAttempted = false;
		}
		else
		{
			if (safeMillisSubtract(now, lastSignalTime) > signalLostTimeoutMs)
			{
#ifdef ENABLE_DEBUG_PRINTING
				Serial.println(F("Signal LOST (Timeout) -> Local Search"));
#endif
				stepperX.stop();
				stepperY.stop();
				changeState(State::LOCAL_SEARCH);
			}
		}
		break;

	case State::LOCAL_SEARCH:
		handleLocalSearch();
		break;

	case State::SEARCH_FAILED:
		if (safeMillisSubtract(now, stateEntryTime) > searchFailDelayMs)
		{
#ifdef ENABLE_DEBUG_PRINTING
			Serial.println(F("Search Failed Timeout -> Retrying Full Search (Sweep X)"));
#endif
			sweepDirectionX = 1;
			changeState(State::SWEEP_X);
		}
		break;

	case State::ERROR:
		break;

	default:
		handleSystemError(F("FATAL: Reached Unknown State!"));
		break;
	}
}

void changeState(State newState)
{
	if (newState == currentState)
		return;

#ifdef ENABLE_DEBUG_PRINTING
	Serial.print(getStateName(currentState));
	Serial.print(F(" -> "));
	Serial.println(getStateName(newState));
#endif

	currentState = newState;
	stateEntryTime = millis();

	if (newState >= State::HOMING_X && newState <= State::LOCAL_SEARCH)
	{
		stallStartTimeX = 0;
		stallStartTimeY = 0;
		lastKnownStepperXPosForStall = stepperX.currentPosition();
		lastKnownStepperYPosForStall = stepperY.currentPosition();
		lastStallCheckTime = stateEntryTime;
	}

	switch (currentState)
	{
	case State::HOMING_X:
		setMotorSpeeds(homingSpeed, homingAcceleration, homingSpeed, homingAcceleration);
		stepperX.moveTo(-((long)stepsPerRevolution * (long)HOMING_OVERSHOOT_MULTIPLIER));
		stepperY.stop();
		break;
	case State::HOMING_Y:
		setMotorSpeeds(homingSpeed, homingAcceleration, homingSpeed, homingAcceleration);
		stepperY.moveTo(-((long)stepsPerRevolution * (long)HOMING_OVERSHOOT_MULTIPLIER));
		stepperX.stop();
		break;
	case State::POST_HOMING_MOVE_OFF_X:
		setMotorSpeeds(homingSpeed, homingAcceleration, homingSpeed, homingAcceleration);
		stepperX.move(postHomingMoveOffSwitchSteps);
		stepperY.stop();
		break;
	case State::POST_HOMING_MOVE_OFF_Y:
		setMotorSpeeds(homingSpeed, homingAcceleration, homingSpeed, homingAcceleration);
		stepperY.move(postHomingMoveOffSwitchSteps);
		stepperX.stop();
		break;
	case State::POST_HOMING_DELAY:
		stepperX.stop();
		stepperY.stop();
		break;
	case State::SWEEP_X:
		setMotorSpeeds(sweepSpeedDefault, sweepAccelerationDefault, sweepSpeedDefault, sweepAccelerationDefault);
		sweepXCyclesCompleted = 0;
		startSweepX();
		break;
	case State::SWEEP_Y:
		setMotorSpeeds(sweepSpeedDefault, sweepAccelerationDefault, sweepSpeedDefault, sweepAccelerationDefault);
		sweepYCyclesCompleted = 0;
		startSweepY();
		break;
	case State::CENTERING:
	case State::TRACKING:
		setMotorSpeeds(trackSpeedDefault, trackAccelerationDefault, trackSpeedDefault, trackAccelerationDefault);
		break;
	case State::LOCAL_SEARCH:
		setMotorSpeeds(localSearchSpeed, localSearchAccel, localSearchSpeed, localSearchAccel);
		currentLocalSearchPhase = 0;
		currentLocalSearchCycle = 0;
#ifdef ENABLE_DEBUG_PRINTING
		Serial.print(F("Entering Local Search. Last Error Int X:"));
		Serial.print(lastErrorX_int);
		Serial.print(F(", Y:"));
		Serial.println(lastErrorY_int);
#endif
		break;
	case State::SEARCH_FAILED:
	case State::ERROR:
		stepperX.stop();
		stepperY.stop();
		break;
	case State::INITIALIZING:
	case State::CHECK_PIN_CONFLICTS:
		break;
	}
	// updateStatusLEDsOptimized(); // REMOVED
}

// ==========================================================================
// ===                       STATE HANDLERS                             ===
// ==========================================================================
void handleHomingStates() // Logic remains same as original
{
	switch (currentState)
	{
	case State::HOMING_X:
		if (limitSwitchActiveX)
		{
#ifdef ENABLE_DEBUG_PRINTING
			Serial.println(F("Homing X: Limit HIT"));
#endif
			stepperX.stop();
			stepperX.setCurrentPosition(searchMinX_steps);
			changeState(State::POST_HOMING_MOVE_OFF_X);
		}
		else if (!stepperX.isRunning())
		{
			handleSystemError(F("Homing X Error: Motor stopped before limit switch."));
		}
		break;

	case State::POST_HOMING_MOVE_OFF_X:
		if (stepperX.distanceToGo() == 0)
		{
#ifdef ENABLE_DEBUG_PRINTING
			Serial.println(F("Homing X: Moved off switch -> Homing Y"));
#endif
			changeState(State::HOMING_Y);
		}
		break;

	case State::HOMING_Y:
		if (limitSwitchActiveY)
		{
#ifdef ENABLE_DEBUG_PRINTING
			Serial.println(F("Homing Y: Limit HIT"));
#endif
			stepperY.stop();
			stepperY.setCurrentPosition(searchMinY_steps);
			changeState(State::POST_HOMING_MOVE_OFF_Y);
		}
		else if (!stepperY.isRunning())
		{
			handleSystemError(F("Homing Y Error: Motor stopped before limit switch."));
		}
		break;

	case State::POST_HOMING_MOVE_OFF_Y:
		if (stepperY.distanceToGo() == 0)
		{
#ifdef ENABLE_DEBUG_PRINTING
			Serial.println(F("Homing Y: Moved off switch -> Post Homing Delay"));
#endif
			changeState(State::POST_HOMING_DELAY);
		}
		break;

	case State::POST_HOMING_DELAY:
		if (safeMillisSubtract(millis(), stateEntryTime) > postHomingPauseMs)
		{
#ifdef ENABLE_DEBUG_PRINTING
			Serial.println(F("Homing Complete -> Starting Full Search (Sweep X)"));
#endif
			sweepDirectionX = 1;
			changeState(State::SWEEP_X);
		}
		break;
	default:
		break;
	}
}

/**
 * @brief Calculates tracking error using photoresistor differences and commands motor movement
 *        using proportional control. Updates `lastErrorX_int`, `lastErrorY_int`.
 */
void handleTrackingLogic()
{
	// sensorReadings[] are populated by readPhotoresistorInputs()
	// Higher analogRead value = more light. Light source attracts.
	int L_val = sensorReadings[SENSOR_L];
	int R_val = sensorReadings[SENSOR_R];
	int U_val = sensorReadings[SENSOR_U];
	int D_val = sensorReadings[SENSOR_D];
	// int C_val = sensorReadings[SENSOR_C]; // Available if needed for more complex logic

	int currentErrorX_int = 0;
	int currentErrorY_int = 0;

	// --- Calculate Raw Difference (Positive error means target is Left/Up) ---
	// X-axis: Positive diff (L > R) means light is to the Left, want to move stepper Right (+steps)
	int x_diff = L_val - R_val;
	// Y-axis: Positive diff (U > D) means light is Up, want to move stepper Down (+steps)
	int y_diff = U_val - D_val;

	// --- Scale Difference to Error Integer ---
	// Tune ANALOG_DIFF_SCALING_FACTOR carefully!
	if (ANALOG_DIFF_SCALING_FACTOR != 0)
	{ // Prevent division by zero
		currentErrorX_int = x_diff / ANALOG_DIFF_SCALING_FACTOR;
		currentErrorY_int = y_diff / ANALOG_DIFF_SCALING_FACTOR;
	}
	else
	{ // Should not happen if configured, but safety
		currentErrorX_int = (x_diff > 0) ? 1 : ((x_diff < 0) ? -1 : 0);
		currentErrorY_int = (y_diff > 0) ? 1 : ((y_diff < 0) ? -1 : 0);
	}

	// --- Store Last Error (for Local Search) ---
	lastErrorX_int = currentErrorX_int;
	lastErrorY_int = currentErrorY_int;

	// --- Apply Proportional Gain (Floating Point) ---
	long stepsToMoveX = (long)((float)currentErrorX_int * Kp_X_float);
	long stepsToMoveY = (long)((float)currentErrorY_int * Kp_Y_float);

	// --- Apply Deadband ---
	if (abs(stepsToMoveX) < trackingDeadband_steps)
		stepsToMoveX = 0;
	if (abs(stepsToMoveY) < trackingDeadband_steps)
		stepsToMoveY = 0;

	// --- Command Movement ---
	if (stepsToMoveX != 0 || stepsToMoveY != 0)
	{
		long currentX = stepperX.currentPosition();
		long currentY = stepperY.currentPosition();
		long targetX = clamp(currentX + stepsToMoveX, searchMinX_steps, searchMaxX_steps);
		long targetY = clamp(currentY + stepsToMoveY, searchMinY_steps, searchMaxY_steps);

		if (targetX != stepperX.targetPosition())
			stepperX.moveTo(targetX);
		if (targetY != stepperY.targetPosition())
			stepperY.moveTo(targetY);
	}
}

void handleLocalSearch() // Logic remains similar to original, uses lastErrorX/Y_int
{
	unsigned long now = millis();

	if (isAnySensorGenerallyActive())
	{
#ifdef ENABLE_DEBUG_PRINTING
		Serial.println(F("Signal RE-ACQUIRED (Local Search) -> Centering"));
#endif
		stepperX.stop();
		stepperY.stop();
		recoveryAttempted = false;
		changeState(State::CENTERING);
		return;
	}

	if (safeMillisSubtract(now, stateEntryTime) > localSearchTimeoutMs)
	{
#ifdef ENABLE_DEBUG_PRINTING
		Serial.println(F("TIMEOUT: Local Search Failed -> Full Search (Sweep X)"));
#endif
		stepperX.stop();
		stepperY.stop();
		sweepDirectionX = 1;
		changeState(State::SWEEP_X);
		return;
	}

	if (stepperX.distanceToGo() == 0 && stepperY.distanceToGo() == 0)
	{
		long currentX = stepperX.currentPosition();
		long currentY = stepperY.currentPosition();
		long targetX = currentX;
		long targetY = currentY;
		bool commandMove = false;

#ifdef ENABLE_DEBUG_PRINTING
		Serial.print(F("Local Search Phase: "));
		Serial.print(currentLocalSearchPhase);
		Serial.print(F(", Cycle/Step: "));
		Serial.println(currentLocalSearchCycle);
#endif

		switch (currentLocalSearchPhase)
		{
		case 0: // Probe
		{
#ifdef ENABLE_DEBUG_PRINTING
			Serial.println(F("  Phase 0: Probing Last Direction"));
#endif
			// lastErrorX_int > 0 means target was Left (L>R), so probe Right (+steps)
			// lastErrorY_int > 0 means target was Up (U>D), so probe Down (+steps)
			targetX = currentX + (lastErrorX_int > 0 ? 1 : (lastErrorX_int < 0 ? -1 : 0)) * localSearchProbeSteps;
			targetY = currentY + (lastErrorY_int > 0 ? 1 : (lastErrorY_int < 0 ? -1 : 0)) * localSearchProbeSteps;

			targetX = clamp(targetX, searchMinX_steps, searchMaxX_steps);
			targetY = clamp(targetY, searchMinY_steps, searchMaxY_steps);
			targetX = clamp(targetX, lastKnownX - localSearchSweepSteps, lastKnownX + localSearchSweepSteps);
			targetY = clamp(targetY, lastKnownY - localSearchSweepSteps, lastKnownY + localSearchSweepSteps);
			commandMove = true;
			currentLocalSearchPhase = 1;
			currentLocalSearchCycle = 0;
			break;
		}
		case 1: // Perpendicular Sweep
		{
#ifdef ENABLE_DEBUG_PRINTING
			Serial.println(F("  Phase 1: Perpendicular Sweep"));
#endif
			int sweepDir = (currentLocalSearchCycle % 2 == 0) ? 1 : -1;
			if (currentLocalSearchCycle < 2)
			{ // Sweep X
				targetX = lastKnownX + sweepDir * localSearchSweepSteps;
				targetY = lastKnownY; // Stay on last known Y for X sweep
			}
			else
			{ // Sweep Y
				targetY = lastKnownY + sweepDir * localSearchSweepSteps;
				targetX = lastKnownX; // Stay on last known X for Y sweep
			}
			targetX = clamp(targetX, searchMinX_steps, searchMaxX_steps);
			targetY = clamp(targetY, searchMinY_steps, searchMaxY_steps);
			commandMove = true;
			currentLocalSearchCycle++;
			if (currentLocalSearchCycle >= 4)
			{
				currentLocalSearchPhase = 2;
				currentLocalSearchCycle = 0;
			}
			break;
		}
		case 2: // Box Pattern
		{
#ifdef ENABLE_DEBUG_PRINTING
			Serial.println(F("  Phase 2: Expanding Box Pattern"));
#endif
			if ((uint16_t)currentLocalSearchCycle >= ((uint16_t)localSearchMaxCycles * LOCAL_SEARCH_BOX_STEPS_PER_LAYER))
			{
#ifdef ENABLE_DEBUG_PRINTING
				Serial.println(F("Local Search Box Pattern Max Cycles Reached -> Full Search"));
#endif
				sweepDirectionX = 1;
				changeState(State::SWEEP_X);
				return;
			}
			long dX = 0, dY = 0;
			int moveScale = (currentLocalSearchCycle / LOCAL_SEARCH_BOX_STEPS_PER_LAYER) + 1;
			switch (currentLocalSearchCycle % LOCAL_SEARCH_BOX_STEPS_PER_LAYER)
			{
			case 0:
				dX = (long)localSearchBoxStepSize * moveScale;
				break; // R
			case 1:
				dY = -(long)localSearchBoxStepSize * moveScale;
				break; // U (photoresistor Y-axis is inverted relative to motor Y for "Up")
			case 2:
				dX = -(long)localSearchBoxStepSize * moveScale;
				break; // L
			case 3:
				dX = -(long)localSearchBoxStepSize * moveScale;
				break; // L
			case 4:
				dY = (long)localSearchBoxStepSize * moveScale;
				break; // D
			case 5:
				dY = (long)localSearchBoxStepSize * moveScale;
				break; // D
			case 6:
				dX = (long)localSearchBoxStepSize * moveScale;
				break; // R
			case 7:
				dX = (long)localSearchBoxStepSize * moveScale;
				break; // R
			}
			targetX = clamp(currentX + dX, searchMinX_steps, searchMaxX_steps);
			targetY = clamp(currentY + dY, searchMinY_steps, searchMaxY_steps);
			commandMove = true;
			currentLocalSearchCycle++;
			break;
		}
		} // end switch

		if (commandMove && (targetX != currentX || targetY != currentY))
		{
#ifdef ENABLE_DEBUG_PRINTING
			Serial.print(F("    Cmd Move -> X:"));
			Serial.print(targetX);
			Serial.print(F(", Y:"));
			Serial.println(targetY);
#endif
			if (targetX != stepperX.targetPosition())
				stepperX.moveTo(targetX);
			if (targetY != stepperY.targetPosition())
				stepperY.moveTo(targetY);
		}
		else if (commandMove)
		{
#ifdef ENABLE_DEBUG_PRINTING
			Serial.println(F("    Local search move resulted in no change."));
#endif
		}
	}
}

void startSweepX() // Logic remains same
{
	long targetX = (sweepDirectionX > 0) ? searchMaxX_steps : searchMinX_steps;
#ifdef ENABLE_DEBUG_PRINTING
	Serial.print(F("Starting X Sweep Leg -> Target: "));
	Serial.println(targetX);
#endif
	stepperX.moveTo(clamp(targetX, searchMinX_steps, searchMaxX_steps));
}

void startSweepY() // Logic remains same
{
	long targetY = (sweepDirectionY > 0) ? searchMaxY_steps : searchMinY_steps;
#ifdef ENABLE_DEBUG_PRINTING
	Serial.print(F("Starting Y Sweep Leg -> Target: "));
	Serial.println(targetY);
#endif
	stepperY.moveTo(clamp(targetY, searchMinY_steps, searchMaxY_steps));
}

// ==========================================================================
// ===                     SETUP & UTILITY FUNCTIONS                      ===
// ==========================================================================
bool setupPinsAndCheckConflicts()
{
#ifdef ENABLE_DEBUG_PRINTING
	Serial.println(F("--- Initializing Pins & Checking Conflicts ---"));
#endif
	// Photoresistor pins are A0-A4 (INPUT by default for analogRead, no explicit pinMode needed)
	// Limit Switches: Input with internal pull-up. MOVED TO D2, D3.
	pinMode(limitSwitchPinX, INPUT_PULLUP); // D2
	pinMode(limitSwitchPinY, INPUT_PULLUP); // D3

	// LEDs REMOVED

	// Stepper pins are configured by AccelStepper.

#ifdef ENABLE_DEBUG_PRINTING
	Serial.println(F("--- Pin Initialization Complete (Review I2C/Analog Pin Warning!) ---"));
#endif
	return true;
}

void printStartupMessage() // Logic remains same
{
#if defined(ENABLE_DEBUG_PRINTING) || defined(ENABLE_TELEMETRY)
	Serial.println(F(""));
	Serial.println(F("-------------------------------------"));
	Serial.print(F(" Photoresistor Tracker System Initializing v"));
	Serial.println(F(SYSTEM_VERSION));
#ifdef DEBUG_BUILD
	Serial.println(F(" Build Mode: DEBUG (Serial Enabled)"));
#else
	Serial.println(F(" Build Mode: PRODUCTION (Serial Disabled)"));
#endif
	Serial.println(F("-------------------------------------"));
	Serial.flush();
#endif
}

void printTuningGuide() // Updated for photoresistors
{
#if defined(ENABLE_DEBUG_PRINTING) || defined(ENABLE_TELEMETRY)
	Serial.println(F("\n--- TUNING REQUIRED - READ CAREFULLY (PHOTORESISTOR VERSION) ---"));
	Serial.println(F("This system WILL NOT WORK reliably without tuning to YOUR specific hardware & light conditions."));
	Serial.println(F("Assumes photoresistors give HIGHER analogRead() value for MORE light."));

	Serial.println(F("\n1. PHOTORESISTOR PARAMETERS (Start Here):"));
	Serial.println(F("   - Goal: Reliable light detection and differential error calculation."));
	Serial.println(F("   - Use Serial Plotter to observe sensorReadings[] values under various light conditions."));
	Serial.print(F("   - PHOTO_LIGHT_THRESHOLD_GENERAL (Currently: "));
	Serial.print(PHOTO_LIGHT_THRESHOLD_GENERAL);
	Serial.println(F("): Min reading for 'isAnySensorGenerallyActive'. Set above ambient noise, below target light."));
	Serial.print(F("   - PHOTO_LIGHT_THRESHOLD_CENTER (Currently: "));
	Serial.print(PHOTO_LIGHT_THRESHOLD_CENTER);
	Serial.println(F("): Similar, for center sensor if specific logic added (currently only for general detection)."));
	Serial.print(F("   - ANALOG_DIFF_SCALING_FACTOR (Currently: "));
	Serial.print(ANALOG_DIFF_SCALING_FACTOR);
	Serial.println(F("): Divides (L-R) or (U-D) difference. Tune so error_int is in a reasonable range (e.g., -20 to +20) for Kp."));
	Serial.println(F("     Too small factor = large error_int = Kp makes huge jumps. Too large factor = small error_int = sluggish."));

	Serial.println(F("\n2. SPEEDS & ACCELERATION:"));
	Serial.println(F("   - Goal: Find fastest reliable speed/accel WITHOUT stalling. Start LOW."));
	Serial.print(F("   - CURRENT DEFAULTS: H="));
	Serial.print(homingSpeed);
	Serial.print(F(" Sw="));
	Serial.print(sweepSpeedDefault);
	Serial.print(F(" Tr="));
	Serial.print(trackSpeedDefault);
	Serial.print(F(" Accel ~"));
	Serial.println(trackAccelerationDefault);

	Serial.println(F("\n3. PROPORTIONAL CONTROL (Kp_X_float, Kp_Y_float, trackingDeadband_steps):"));
	Serial.println(F("   - Kp (Gain): Tune Kp_X_float / Kp_Y_float. Determines reaction strength. START LOW (e.g., 5.0-20.0)."));
	Serial.println(F("     - TOO LOW: Sluggish. -> INCREASE Kp_float."));
	Serial.println(F("     - TOO HIGH: Overshoots, oscillates. -> DECREASE Kp_float."));
	Serial.println(F("   - Deadband: Zone (steps) around center where no correction occurs."));
	Serial.print(F("   - CURRENT DEFAULTS: Kp_X_float="));
	Serial.print(Kp_X_float);
	Serial.print(F(", Kp_Y_float="));
	Serial.print(Kp_Y_float);
	Serial.print(F(", Deadband="));
	Serial.println(trackingDeadband_steps);

	Serial.println(F("\n4. STALL DETECTION & OTHER TIMINGS: Review if issues arise."));
	Serial.println(F("\n--- End Tuning Guidance ---"));
	Serial.flush();
#endif
}

/**
 * @brief Reads photoresistor values and limit switch states.
 *        Updates global `sensorReadings[]`, `sensorActive[]`, `limitSwitchActiveX/Y`.
 */
void readPhotoresistorInputs()
{
	unsigned long now = millis();

	// --- Read Photoresistors ---
	for (int i = 0; i < NUM_SENSORS; i++)
	{
		sensorReadings[i] = analogRead(photoresistorPins[i]);
		sensorActive[i] = (sensorReadings[i] > PHOTO_LIGHT_THRESHOLD_GENERAL);
	}

	// --- Read Limit Switches (on PORTD: D2, D3) ---
	uint8_t currentPinStatesD_limits = PIND; // Read PORTD for limit switches

	// Check for Changes & Reset Debounce Timers for Limit Switches
	bool limitsChanged = ((lastRawPinStatesD ^ currentPinStatesD_limits) & (limitSwitchMaskX | limitSwitchMaskY));

	if (limitsChanged)
	{
		lastDebounceTimeLimits = now;
	}
	lastRawPinStatesD = currentPinStatesD_limits; // Update last raw state FOR LIMITS

	// Apply Debounce Logic for Limit Switches
	if (safeMillisSubtract(now, lastDebounceTimeLimits) > limitSwitchDebounceMs)
	{
		limitSwitchActiveX = !(currentPinStatesD_limits & limitSwitchMaskX);
		limitSwitchActiveY = !(currentPinStatesD_limits & limitSwitchMaskY);
	}
}

/**
 * @brief Checks if any photoresistor is reading above the general threshold.
 */
inline bool isAnySensorGenerallyActive()
{
	for (uint8_t i = 0; i < NUM_SENSORS; i++)
	{
		if (sensorActive[i])
			return true; // Uses the pre-calculated sensorActive based on threshold
	}
	return false;
}

/**
 * @brief Checks if the center photoresistor is reading above the general threshold.
 */
inline bool isCenterSensorGenerallyActive()
{
	return sensorActive[SENSOR_C]; // Uses the pre-calculated sensorActive
}

// updateStatusLEDsOptimized() REMOVED

void checkStateTimeouts() // Logic remains same
{
	unsigned long currentTimeout = 0;
	switch (currentState)
	{
	case State::HOMING_X:
	case State::HOMING_Y:
	case State::POST_HOMING_MOVE_OFF_X:
	case State::POST_HOMING_MOVE_OFF_Y:
		currentTimeout = homingTimeoutMs;
		break;
	case State::SWEEP_X:
	case State::SWEEP_Y:
	case State::CENTERING:
		currentTimeout = stateTimeoutMs;
		break;
	default:
		return;
	}
	if (currentTimeout > 0 && safeMillisSubtract(millis(), stateEntryTime) > currentTimeout)
	{
#ifdef ENABLE_DEBUG_PRINTING
		Serial.print(F("TIMEOUT! State: "));
		Serial.print(getStateName(currentState));
		Serial.print(F(" exceeded "));
		Serial.print(currentTimeout);
		Serial.println(F(" ms"));
#endif
		handleSystemError(F("Safety Timeout: State duration exceeded limit."));
	}
}

void checkMotorStall() // Logic remains same
{
	unsigned long now = millis();
	if (safeMillisSubtract(now, lastStallCheckTime) < stallCheckIntervalMs)
		return;
	lastStallCheckTime = now;

	bool stallDetected = false;
	const __FlashStringHelper *stallReason = nullptr;

	if (stepperX.isRunning())
	{
		long currentXPos = stepperX.currentPosition();
		long deltaX = currentXPos - lastKnownStepperXPosForStall;
		if (deltaX < 0)
			deltaX = -deltaX;
		if (deltaX < stallPositionTolerance)
		{
			if (stallStartTimeX == 0)
				stallStartTimeX = now;
			else if (safeMillisSubtract(now, stallStartTimeX) > stallTimeoutThresholdMs)
			{
				stallDetected = true;
				stallReason = F("Stall Detected: Motor X stuck.");
			}
		}
		else
		{
			stallStartTimeX = 0;
			lastKnownStepperXPosForStall = currentXPos;
		}
	}
	else
	{
		stallStartTimeX = 0;
		lastKnownStepperXPosForStall = stepperX.currentPosition();
	}

	if (!stallDetected && stepperY.isRunning())
	{
		long currentYPos = stepperY.currentPosition();
		long deltaY = currentYPos - lastKnownStepperYPosForStall;
		if (deltaY < 0)
			deltaY = -deltaY;
		if (deltaY < stallPositionTolerance)
		{
			if (stallStartTimeY == 0)
				stallStartTimeY = now;
			else if (safeMillisSubtract(now, stallStartTimeY) > stallTimeoutThresholdMs)
			{
				stallDetected = true;
				stallReason = F("Stall Detected: Motor Y stuck.");
			}
		}
		else
		{
			stallStartTimeY = 0;
			lastKnownStepperYPosForStall = currentYPos;
		}
	}
	else
	{
		stallStartTimeY = 0;
		lastKnownStepperYPosForStall = stepperY.currentPosition();
	}

	if (stallDetected)
		handleSystemError(stallReason);
}

void handleSystemError(const __FlashStringHelper *reason) // Logic remains same
{
#if defined(ENABLE_DEBUG_PRINTING) || defined(ENABLE_TELEMETRY)
	Serial.println(F("--- System Error Detected ---"));
	Serial.println(reason);
#endif
	bool canAttemptRecovery = (currentState >= State::SWEEP_X && currentState <= State::LOCAL_SEARCH);
	if (!recoveryAttempted && canAttemptRecovery)
	{
#if defined(ENABLE_DEBUG_PRINTING) || defined(ENABLE_TELEMETRY)
		Serial.println(F("Attempting recovery: Stopping motors and restarting full search..."));
#endif
		recoveryAttempted = true;
		stepperX.stop();
		stepperY.stop();
		sweepDirectionX = 1;
		changeState(State::SWEEP_X);
		return;
	}
#if defined(ENABLE_DEBUG_PRINTING) || defined(ENABLE_TELEMETRY)
	if (recoveryAttempted && canAttemptRecovery)
		Serial.println(F("Recovery already attempted but failed. Proceeding to fatal error."));
	else if (!canAttemptRecovery)
		Serial.println(F("Error occurred during non-recoverable state. Proceeding to fatal error."));
#endif
	indicateFatalError(reason);
}

void indicateFatalError(const __FlashStringHelper *message) // Modified for no LEDs
{
#if defined(ENABLE_DEBUG_PRINTING) || defined(ENABLE_TELEMETRY)
	Serial.println(F("\n!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!"));
	Serial.println(F("          >>> FATAL SYSTEM ERROR <<<"));
	Serial.println(F("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!"));
	Serial.print(F("REASON: "));
	Serial.println(message);
	Serial.print(F("STATE AT ERROR: "));
	Serial.println(getStateName(currentState));
	Serial.print(F("TIME IN STATE (ms): "));
	Serial.println(safeMillisSubtract(millis(), stateEntryTime));
	Serial.print(F("Position X: "));
	Serial.print(stepperX.currentPosition());
	Serial.print(F(" Position Y: "));
	Serial.println(stepperY.currentPosition());
	Serial.print(F("Photoresistors Raw [L R U D C]: "));
	for (int i = 0; i < NUM_SENSORS; i++)
	{
		Serial.print(sensorReadings[i]);
		Serial.print(i == NUM_SENSORS - 1 ? "" : " ");
	}
	Serial.println();
	Serial.print(F("Limits Active [X Y]: "));
	Serial.print(limitSwitchActiveX ? '1' : '0');
	Serial.print(" ");
	Serial.println(limitSwitchActiveY ? '1' : '0');
	Serial.println(F("System Halted. Reset required."));
	Serial.println(F("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n"));
	Serial.flush();
#endif
	noInterrupts();
	currentState = State::ERROR;
	stepperX.disableOutputs();
	stepperY.disableOutputs();
	while (true)
	{
		// Halt CPU. No LEDs to blink.
		delay(1000); // Simple delay in halt loop
	}
}

inline unsigned long safeMillisSubtract(unsigned long current, unsigned long previous) // Logic remains same
{
	if (current >= previous)
		return current - previous;
	else
		return (ULONG_MAX - previous) + current + 1;
}

inline long clamp(long val, long minVal, long maxVal) // Logic remains same
{
	if (val < minVal)
		return minVal;
	if (val > maxVal)
		return maxVal;
	return val;
}

void setMotorSpeeds(float speedX, float accelX, float speedY, float accelY) // Logic remains same
{
	stepperX.setMaxSpeed(speedX);
	stepperX.setAcceleration(accelX);
	stepperY.setMaxSpeed(speedY);
	stepperY.setAcceleration(accelY);
}

const char *getStateName(State state) // Logic remains same
{
	strcpy_P(stateNameBuffer, (char *)pgm_read_word(&(stateNames[static_cast<uint8_t>(state)])));
	return stateNameBuffer;
}

void printTelemetry() // Updated for photoresistor active state
{
#ifdef ENABLE_TELEMETRY
	Serial.print(F("TLM: St:"));
	Serial.print(getStateName(currentState));
	Serial.print(F(" P:[X:"));
	Serial.print(stepperX.currentPosition());
	Serial.print(F(", Y:"));
	Serial.print(stepperY.currentPosition());
	Serial.print(F("]"));
	Serial.print(F(" PhRS Actv[LRCUD]: "));			  // PhotoResistor Active
	Serial.print(sensorActive[SENSOR_L] ? '1' : '0'); // Using the boolean sensorActive array
	Serial.print(sensorActive[SENSOR_R] ? '1' : '0');
	Serial.print(sensorActive[SENSOR_C] ? '1' : '0'); // Note: C is in middle of L R U D for this print
	Serial.print(sensorActive[SENSOR_U] ? '1' : '0');
	Serial.print(sensorActive[SENSOR_D] ? '1' : '0');
	Serial.print(F("] PhRS Raw[C]:"));
	Serial.print(sensorReadings[SENSOR_C]); // Example raw value
	Serial.println();
#endif
}